import argparse
import asyncio
import dataclasses
import json
import random
import time
from collections import defaultdict
from datetime import datetime, timedelta

import numpy as np
from websockets import connect

from cooperative_cuisine.action import ActionType, InterActionData, Action
from cooperative_cuisine.utils import custom_asdict_factory

TIME_TO_STOP_ACTION = 3.0


async def agent():
    parser = argparse.ArgumentParser("Random agent")
    parser.add_argument("--uri", type=str)
    parser.add_argument("--player_id", type=str)
    parser.add_argument("--player_hash", type=str)
    parser.add_argument("--step_time", type=float, default=0.5)

    args = parser.parse_args()

    async with connect(args.uri) as websocket:
        await websocket.send(
            json.dumps({"type": "ready", "player_hash": args.player_hash})
        )
        await websocket.recv()

        ended = False

        counters = None

        player_info = {}
        current_agent_pos = None
        interaction_counter = None

        last_interacting = False
        last_interact_progress = None

        threshold = datetime.max

        task_type = None
        task_args = None

        started_interaction = False
        still_interacting = False
        current_nearest_counter_id = None

        while not ended:
            time.sleep(args.step_time)
            await websocket.send(
                json.dumps({"type": "get_state", "player_hash": args.player_hash})
            )
            state = json.loads(await websocket.recv())
            if not state["all_players_ready"]:
                continue

            if counters is None:
                counters = defaultdict(list)
                for counter in state["counters"]:
                    counters[counter["type"]].append(counter)

            for player in state["players"]:
                if player["id"] == args.player_id:
                    player_info = player
                    current_agent_pos = player["pos"]
                    if player["current_nearest_counter_id"]:
                        if (
                            current_nearest_counter_id
                            != player["current_nearest_counter_id"]
                        ):
                            for counter in state["counters"]:
                                if (
                                    counter["id"]
                                    == player["current_nearest_counter_id"]
                                ):
                                    interaction_counter = counter
                                    current_nearest_counter_id = player[
                                        "current_nearest_counter_id"
                                    ]
                                    break
                    if last_interacting:
                        if (
                            not interaction_counter
                            or not interaction_counter["occupied_by"]
                            or isinstance(interaction_counter["occupied_by"], list)
                            or (
                                interaction_counter["occupied_by"][
                                    "progress_percentage"
                                ]
                                == 1.0
                            )
                        ):
                            last_interacting = False
                            last_interact_progress = None
                    else:
                        if (
                            interaction_counter
                            and interaction_counter["occupied_by"]
                            and not isinstance(interaction_counter["occupied_by"], list)
                        ):
                            if (
                                last_interact_progress
                                != interaction_counter["occupied_by"][
                                    "progress_percentage"
                                ]
                            ):
                                last_interact_progress = interaction_counter[
                                    "occupied_by"
                                ]["progress_percentage"]
                                last_interacting = True

                    break

            if task_type:
                if threshold < datetime.now():
                    print(
                        args.player_hash, args.player_id, "---Threshold---Too long---"
                    )
                    task_type = None
                match task_type:
                    case "GOTO":
                        diff = np.array(task_args) - np.array(current_agent_pos)
                        dist = np.linalg.norm(diff)
                        if dist > 1.2:
                            if dist != 0:
                                await websocket.send(
                                    json.dumps(
                                        {
                                            "type": "action",
                                            "action": dataclasses.asdict(
                                                Action(
                                                    args.player_id,
                                                    ActionType.MOVEMENT,
                                                    (diff / dist).tolist(),
                                                    args.step_time + 0.01,
                                                ),
                                                dict_factory=custom_asdict_factory,
                                            ),
                                            "player_hash": args.player_hash,
                                        }
                                    )
                                )
                                await websocket.recv()
                        else:
                            task_type = None
                            task_args = None
                    case "INTERACT":
                        if not started_interaction or (
                            still_interacting and interaction_counter
                        ):
                            if not started_interaction:
                                started_interaction = True

                            still_interacting = True
                            await websocket.send(
                                json.dumps(
                                    {
                                        "type": "action",
                                        "action": dataclasses.asdict(
                                            Action(
                                                args.player_id,
                                                ActionType.INTERACT,
                                                InterActionData.START,
                                            ),
                                            dict_factory=custom_asdict_factory,
                                        ),
                                        "player_hash": args.player_hash,
                                    }
                                )
                            )
                            await websocket.recv()
                        else:
                            still_interacting = False
                            started_interaction = False
                            task_type = None
                            task_args = None
                    case "PUT":
                        await websocket.send(
                            json.dumps(
                                {
                                    "type": "action",
                                    "action": dataclasses.asdict(
                                        Action(
                                            args.player_id,
                                            ActionType.PUT,
                                            "pickup",
                                        ),
                                        dict_factory=custom_asdict_factory,
                                    ),
                                    "player_hash": args.player_hash,
                                }
                            )
                        )
                        await websocket.recv()
                        task_type = None
                        task_args = None
                    case None:
                        ...

            if not task_type:
                task_type = random.choice(["GOTO", "PUT", "INTERACT"])
                threshold = datetime.now() + timedelta(seconds=TIME_TO_STOP_ACTION)
                if task_type == "GOTO":
                    counter_type = random.choice(list(counters.keys()))
                    task_args = random.choice(counters[counter_type])["pos"]
                    print(args.player_hash, args.player_id, task_type, counter_type)
                else:
                    print(args.player_hash, args.player_id, task_type)
                    task_args = None

            ended = state["ended"]


if __name__ == "__main__":
    asyncio.run(agent())
