"""
Record events in jsonl-files.

- `USER_LOG_DIR` corresponds to the log directory of your system for programs. See [platformdirs](
https://pypi.org/project/platformdirs/) -> `user_log_dir`.
- `ROOT_DIR` to the location of the cooperative_cuisine.
- `ENV_NAME` to the name of the environment.

```yaml
extra_setup_functions:
  json_states:
    func: !!python/name:cooperative_cuisine.hooks.hooks_via_callback_class ''
    kwargs:
      hooks: [ json_state ]
      callback_class: !!python/name:cooperative_cuisine.recording.FileRecorder ''
      callback_class_kwargs:
        log_path: USER_LOG_DIR/ENV_NAME/json_states.jsonl
  actions:
    func: !!python/name:cooperative_cuisine.hooks.hooks_via_callback_class ''
    kwargs:
      hooks: [ pre_perform_action ]
      callback_class: !!python/name:cooperative_cuisine.recording.FileRecorder ''
      callback_class_kwargs:
        log_path: USER_LOG_DIR/ENV_NAME/LOG_RECORD_NAME.jsonl
  random_env_events:
    func: !!python/name:cooperative_cuisine.hooks.hooks_via_callback_class ''
    kwargs:
      hooks: [ order_duration_sample, plate_out_of_kitchen_time ]
      callback_class: !!python/name:cooperative_cuisine.recording.FileRecorder ''
      callback_class_kwargs:
        log_path: USER_LOG_DIR/ENV_NAME/LOG_RECORD_NAME.jsonl
        add_hook_ref: true
  env_configs:
    func: !!python/name:cooperative_cuisine.hooks.hooks_via_callback_class ''
    kwargs:
      hooks: [ env_initialized, item_info_config ]
      callback_class: !!python/name:cooperative_cuisine.recording.FileRecorder ''
      callback_class_kwargs:
        log_path: USER_LOG_DIR/ENV_NAME/LOG_RECORD_NAME.jsonl
        add_hook_ref: true
```
"""
import json
import logging
import os
import traceback
from pathlib import Path

from cooperative_cuisine.environment import Environment
from cooperative_cuisine.hooks import HookCallbackClass
from cooperative_cuisine.utils import NumpyAndDataclassEncoder, expand_path

log = logging.getLogger(__name__)
"""The logger for this module."""


class FileRecorder(HookCallbackClass):
    """
    Class: FileRecorder

    This class is responsible for recording data to a file.

    Attributes:
        name (str): The name of the recorder.
        env (Environment): The environment instance.
        log_path (str): The path to the log file. Default value is "USER_LOG_DIR/ENV_NAME/LOG_RECORD_NAME.jsonl".
        add_hook_ref (bool): Indicates whether to add a hook reference to the recorded data. Default value is False.

    """

    def __init__(
        self,
        name: str,
        env: Environment,
        log_path: str = "USER_LOG_DIR/ENV_NAME/LOG_RECORD_NAME.jsonl",
        add_hook_ref: bool = False,
        **kwargs,
    ):
        super().__init__(name, env, **kwargs)
        self.add_hook_ref = add_hook_ref
        """If the name of the hook (the reference) should be included in the recording."""
        log_path = log_path.replace("LOG_RECORD_NAME", name)
        log_path = Path(expand_path(log_path, env_name=env.env_name))
        self.log_path = log_path
        """The path to the recording file."""
        log.info(f"Recorder record for {name} in file://{log_path}")
        os.makedirs(log_path.parent, exist_ok=True)

    def __call__(self, hook_ref: str, env: Environment, **kwargs):
        try:
            record = (
                json.dumps(
                    {
                        "env_time": env.env_time.isoformat(),
                        **kwargs,
                        **({"hook_ref": hook_ref} if self.add_hook_ref else {}),
                    },
                    cls=NumpyAndDataclassEncoder,
                )
                + "\n"
            )
            with open(self.log_path, "a") as log_file:
                log_file.write(record)
        except TypeError as e:
            traceback.print_exception(e)
            log.info(f"Not JSON serializable Record {kwargs}")
