from datetime import timedelta

import numpy as np
import pytest

from cooperative_cuisine import ROOT_DIR
from cooperative_cuisine.action import ActionType, InterActionData, Action
from cooperative_cuisine.counters import Counter, CuttingBoard
from cooperative_cuisine.environment import (
    Environment,
)
from cooperative_cuisine.game_server import PlayerRequestType
from cooperative_cuisine.hooks import Hooks
from cooperative_cuisine.items import Item, ItemInfo, ItemType
from cooperative_cuisine.server_results import (
    PlayerInfo,
    CreateEnvResult,
    PlayerRequestResult,
)
from cooperative_cuisine.state_representation import (
    StateRepresentation,
    create_json_schema,
)
from cooperative_cuisine.utils import create_init_env_time

layouts_folder = ROOT_DIR / "configs" / "layouts"
environment_config_path = ROOT_DIR / "configs" / "environment_config.yaml"
layout_path = ROOT_DIR / "configs" / "layouts" / "basic.layout"
layout_empty_path = ROOT_DIR / "configs" / "layouts" / "basic.layout"
item_info_path = ROOT_DIR / "configs" / "item_info.yaml"

# TODO: TESTs are in absolute pixel coordinates still.


@pytest.fixture(autouse=True)
def test_file_availability():
    assert layouts_folder.is_dir(), "layouts folder does not exists"
    assert environment_config_path.is_file(), "environment config file does not exists"
    assert layout_path.is_file(), "layout config file does not exists"
    assert layout_empty_path.is_file(), "layout empty config file does not exists"
    assert item_info_path.is_file(), "item info config file does not exists"


@pytest.fixture
def env_config():
    with open(environment_config_path, "r") as file:
        env_config = file.read()

    return env_config


@pytest.fixture
def layout_config():
    with open(layout_path, "r") as file:
        layout = file.read()
    return layout
    env.add_player("0")


@pytest.fixture
def layout_empty_config():
    with open(layout_path, "r") as file:
        layout = file.read()
    return layout


@pytest.fixture
def item_info():
    with open(item_info_path, "r") as file:
        item_info = file.read()
    return item_info


def test_player_registration(env_config, layout_config, item_info):
    env = Environment(env_config, layout_config, item_info, as_files=False)

    env.add_player("1")
    assert len(env.players) == 1, "Wrong number of players"

    env.add_player("2")
    assert len(env.players) == 2, "Wrong number of players"

    with pytest.raises(ValueError):
        env.add_player("2")


def test_movement(env_config, layout_empty_config, item_info):
    env = Environment(env_config, layout_empty_config, item_info, as_files=False)
    player_name = "1"
    start_pos = np.array([3, 4])
    env.add_player(player_name, start_pos)
    env.movement.player_movement_speed = 1
    move_direction = np.array([1, 0])
    move_action = Action(player_name, ActionType.MOVEMENT, move_direction, duration=0.1)
    do_moves_number = 3
    for i in range(do_moves_number):
        env.perform_action(action=move_action)
        env.step(timedelta(seconds=0.1))

    expected = start_pos + do_moves_number * (
        move_direction * env.movement.player_movement_speed * move_action.duration
    )
    assert np.isclose(
        np.linalg.norm(expected - env.players[player_name].pos), 0
    ), "Performed movement do not move the player as expected."


def test_player_movement_speed(env_config, layout_empty_config, item_info):
    env = Environment(env_config, layout_empty_config, item_info, as_files=False)
    player_name = "1"
    start_pos = np.array([3, 4])
    env.add_player(player_name, start_pos)
    env.movement.player_movement_speed = 2
    move_direction = np.array([1, 0])
    move_action = Action(player_name, ActionType.MOVEMENT, move_direction, duration=0.1)
    do_moves_number = 3
    for i in range(do_moves_number):
        env.perform_action(action=move_action)
        env.step(timedelta(seconds=0.1))

    expected = start_pos + do_moves_number * (
        move_direction * env.movement.player_movement_speed * move_action.duration
    )

    assert np.isclose(
        np.linalg.norm(expected - env.players[player_name].pos), 0
    ), "Performed movement do not move the player as expected."

    assert StateRepresentation.model_validate_json(
        json_data=env.get_json_state(player_id="1")
    ), "json state does not match expected StateRepresentation."


def test_player_reach(env_config, layout_empty_config, item_info):
    env = Environment(env_config, layout_empty_config, item_info, as_files=False)

    counter_pos = np.array([2, 2])
    counter = Counter(pos=counter_pos, hook=Hooks(env))
    env.overwrite_counters([counter])
    env.add_player("1", np.array([2, 4]))
    env.movement.player_movement_speed = 1
    player = env.players["1"]
    assert not player.can_reach(counter), "Player is too far away."

    do_moves_number = 30
    for i in range(do_moves_number):
        move_action = Action("1", ActionType.MOVEMENT, np.array([0, -1]), duration=1)
        env.perform_action(move_action)
        env.step(passed_time=timedelta(seconds=1))
    assert player.can_reach(counter), "Player can reach counter?"


def test_pickup(env_config, layout_config, item_info):
    env = Environment(env_config, layout_config, item_info, as_files=False)

    counter_pos = np.array([2, 2])
    counter = Counter(pos=counter_pos, hook=Hooks(env))
    counter.occupied_by = Item(name="Tomato", item_info=None)
    env.overwrite_counters([counter])

    env.add_player("1", np.array([2, 3]))
    player = env.players["1"]
    env.movement.player_movement_speed = 1

    move_down = Action("1", ActionType.MOVEMENT, np.array([0, -1]), duration=1)
    move_up = Action("1", ActionType.MOVEMENT, np.array([0, 1]), duration=1)
    pick = Action("1", ActionType.PUT, "pickup")

    env.perform_action(move_down)
    env.step(timedelta(seconds=1))
    assert player.can_reach(counter), "Player can reach counter?"

    env.perform_action(pick)
    assert player.holding is not None, "Player should have picked up tomato."
    assert player.holding.name == "Tomato", "Should be tomato."

    for _ in range(5):
        env.perform_action(move_up)
        env.step(timedelta(seconds=1))

    env.perform_action(pick)
    assert (
        player.holding is not None
    ), "Player should be too far away to put tomato down."

    for _ in range(4):
        env.perform_action(move_down)
        env.step(timedelta(seconds=1))
        env.perform_action(move_down)
        env.step(timedelta(seconds=1))
        env.perform_action(move_down)
        env.step(timedelta(seconds=1))
        env.perform_action(move_down)
        env.step(timedelta(seconds=1))

    env.perform_action(pick)

    assert player.holding is None, "Player should have put tomato down."
    assert (
        counter.occupied_by is not None and counter.occupied_by.name == "Tomato"
    ), "Tomato should be here now."


def test_processing(env_config, layout_config, item_info):
    env = Environment(env_config, layout_config, item_info, as_files=False)
    counter_pos = np.array([2, 2])
    counter = CuttingBoard(
        pos=counter_pos,
        hook=Hooks(env),
        transitions={
            "ChoppedTomato": ItemInfo(
                name="ChoppedTomato",
                seconds=0.5,
                equipment=ItemInfo(name="CuttingBoard", type=ItemType.Equipment),
                type=ItemType.Ingredient,
                needs=["Tomato"],
            )
        },
    )
    env.counters.append(counter)

    tomato = Item(name="Tomato", item_info=None)
    env.add_player("1", np.array([2, 3]))
    player = env.players["1"]
    env.movement.player_movement_speed = 1
    player.holding = tomato

    move = Action("1", ActionType.MOVEMENT, np.array([0, -1]), duration=1)
    pick = Action("1", ActionType.PUT, "pickup")

    env.perform_action(move)
    env.step(timedelta(seconds=1))
    env.perform_action(pick)

    hold_down = Action("1", ActionType.INTERACT, InterActionData.START)
    env.perform_action(hold_down)

    assert tomato.name != "ChoppedTomato", "Tomato is not finished yet."

    env.step(timedelta(seconds=1))

    assert tomato.name == "ChoppedTomato", "Tomato should be finished."

    button_up = Action("1", ActionType.INTERACT, InterActionData.STOP)
    env.perform_action(button_up)
    env.perform_action(pick)
    assert player.holding.name == "ChoppedTomato", "Tomato should be finished."


def test_time_passed():
    np.random.seed(42)
    env = Environment(
        ROOT_DIR / "configs" / "environment_config.yaml",
        layouts_folder / "basic.layout",
        ROOT_DIR / "configs" / "item_info.yaml",
    )
    env.add_player("0")
    env.reset_env_time()
    passed_time = timedelta(seconds=10)
    env.step(passed_time)
    assert (
        env.env_time == create_init_env_time() + passed_time
    ), "Env time needs to be updated via the step function"

    passed_time_2 = timedelta(seconds=12)
    env.step(passed_time_2)
    assert (
        env.env_time == create_init_env_time() + passed_time + passed_time_2
    ), "Env time needs to be updated via the step function"


def test_time_limit():
    np.random.seed(42)
    env = Environment(
        ROOT_DIR / "configs" / "environment_config.yaml",
        layouts_folder / "basic.layout",
        ROOT_DIR / "configs" / "item_info.yaml",
    )
    env.add_player("0")

    env.reset_env_time()

    assert not env.game_ended, "Game has not ended yet"

    passed_time = timedelta(seconds=10)
    env.step(passed_time)

    assert not env.game_ended, "Game has not ended yet"

    passed_time_2 = timedelta(
        seconds=(env.env_time_end - env.start_time).total_seconds()
    )
    env.step(passed_time_2)

    assert env.game_ended, "Game has ended now."


def test_json_schema():
    assert isinstance(create_json_schema(), dict)


def test_server_result_definition():
    plater_info = PlayerInfo(client_id="123", player_hash="234567890", player_id="0")
    CreateEnvResult(env_id="123344", player_info={"0": plater_info}, recipe_graphs=[])
    PlayerRequestResult(
        request_type=PlayerRequestType.READY,
        status=200,
        msg="123",
        player_hash="1234324",
    )
